# Module of Foswiki - The Free and Open Source Wiki, http://foswiki.org/
#
# This module is a derivative of Foswiki::Contrib::LdapContrib (version 5.04),
# which has the following copyright. This derivative work has been funded
# through the University of Wisconsin-Milwaukee, on behalf of the LIGO
# Scientific Collaboration. Coding done by Mike Grady, Unicon, Inc., based
# on requirements provided by the funders.
#
# Copyright (C) 2006-2012 Michael Daum http://michaeldaumconsulting.com
# Portions Copyright (C) 2006 Spanlink Communications
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read LICENSE in the root of this distribution.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# As per the GPL, removal of this notice is prohibited.

package Foswiki::Contrib::ShibLdapContrib;

use strict;
use warnings;

use Net::LDAP qw(LDAP_REFERRAL);
use URI::ldap;
use Net::LDAP::Constant qw(LDAP_SUCCESS LDAP_SIZELIMIT_EXCEEDED LDAP_CONTROL_PAGED);
use Net::LDAP::Extension::SetPassword;
use DB_File;
use Encode ();

use Foswiki::Func ();
use Foswiki::Plugins ();

use version; our $VERSION = version->declare("v1.0.0");
our $RELEASE = '06 May 2014';
# derived from and including most of LdapContrib: our $VERSION = '5.04'; our $RELEASE = '5.04';

# One-line description of the module
our $SHORTDESCRIPTION = 'Provide support for both authentication and authorization (group membership) information coming from Shibboleth (SAML), while retaining support for LDAP-based groups';

our %sharedShibLdapContrib;

=pod

---+ Foswiki::Contrib::ShibLdapContrib

General Group/LDAP services module that has been modified to work with Shibboleth. 
This class encapsulates the platform-specific means to integrate an LDAP directory service
mand group memberships supplied thru Shibboleth/SAML.  Used by Foswiki::LoginManager::ShibLdapApacheLogin
for authentication, and Foswiki::Users::ShibLdapUserMapping for group definitions.
Note that, unlike LdapContrib that this is derived from, "person objects" are not expected to
be read in and used from a local LDAP source, but "group objects" will be (if configured to be).
This module allows for group memberships to be consumed from both an ENVVAR populated by the
Shib SP, and/or group memberships from whatever local LDAP groups we consume groups. But note
that since login is based on users logging in based on their home/chosen IdP, those users are
not assumed to have a person record in the local LDAP, and thus they do not have a local LDAP
DN. So DN-valued group memberships are really not useful in this case, you need to have groups
with a "hasMember" attribute that carries the login name (the login name that users will end up
with after whatever username manipulation you have configured to happen) of the member users.
p.s. Note we haven't removed the ability to read in a user tree from a local LDAP server,
if for some reason somebody needed such. Not sure how those could be useful, but the ability
is left here for that.

 Note that this variation of LdapContrib still retains the Cache file from that, for
 caching information read from a LDAP server. But that, given one is relying on identities
 coming from Shib, it only makes sense to cache Group information, and to use Group objects
 that contain a member-type attribute that is <b>NOT</b> a DN, because these users
 won't/don't need to have a "local identity" in that LDAP server. And the loginName and
 group membership(s) delivered by Shib will not be written into that Cache, because:
  - they'd get deleted as soon as the Cache is refreshed the next time
  - it would cause writes to the Cache with almost every new user interaction
    with Foswiki, greatly raising the possibility of clashes
  - this will perform better by not trying to add the Shib info into the cache
  - the Shib info is in the environment with every interaction

---++ Typical usage
<verbatim>
my $ldap = new Foswiki::Contrib::ShibLdapContrib;

$ldap->search(
  filter => 'mail=*@gmx*',
  callback => sub {
    my ($ldap, $entry) = @_;
    # process entry
  } 
);
my $errorMsg = $ldap->getError();

my $count = $result->count();

my @entries = $result->sorted('sn');
my $entry = $result->entry(0);

my $value = $entry->get_value('cn');
my @emails = $entry->get_value('mail');
</verbatim>

---++ Cache storage format

The cache stores a series of key-value pairs in a DB_File. The following
keys are used:

   * WIKINAMES - list of all wikiNames
   * LOGINNAMES - list of all loginNames
   * GROUPS - list of all groups
   * UNKWNUSERS - list of all usernames that could not be found in LDAP (to avoid future LDAP lookups in case caching is OFF)
   * UNKWNGROUPS - list of all group names that could not be found in LDAP (to avoid future LDAP lookups in case caching is OFF)
   * GROUPS::$groupName - list of all loginNames in group groupName (membership)
   * GROUP2UNCACHEDMEMBERSDN::$groupName - list of all DNs (when in memberIndirection mode) that could not be resolved to a user or group existing in the cache when $groupName was retreived from LDAP
   * EMAIL2U::$emailAddr - stores the loginName of an emailAddr
   * U2EMAIL::$loginName - stores the emailAddr of a loginName 
   * U2W::$loginName - stores the wikiName of a  loginName
   * W2U::$wikiName - stores the loginName of a wikiName
   * DN2U::$dn - stores the loginName of a distinguishedName
   * U2DN::$loginName - stores the distinguishedName of a loginName

=cut

=pod

---++ writeDebug($msg) 

Static Method to write a debug messages. 

=cut

sub writeDebug {
  Foswiki::Func::writeDebug("ShibLdapContrib - $_[0]\n") if $Foswiki::cfg{ShibLdap}{Debug};
}

=pod

---++ writeWarning($msg) 

Static Method to write a warning messages. 

=cut

sub writeWarning {
  print STDERR "- ShibLdapContrib - WARNING: $_[0]\n";
}

=pod

---++ new($session, host=>'...', base=>'...', ...) -> $ldap

Construct a new Foswiki::Contrib::ShibLdapContrib object

Possible options are:
   * host: ip address (or hostname) 
   * base: the base DN to use in searches
   * port: port address used when binding to the LDAP server
   * version: protocol version 
   * userBase: list of sub-trees containing user accounts
   * groupBase: list of sub-trees containing group definitions
   * loginAttribute: user login name attribute
   * loginFilter: filter to be used to find login accounts
   * groupAttribute: the group name attribute 
   * groupFilter: filter to be used to find groups
   * memberAttribute: the attribute that should be used to collect group members
   * innerGroupAttribute: the attribute that should be used to collect inner groups of a group
   * bindDN: the dn to use when binding to the LDAP server
   * bindPassword: the password used when binding to the LDAP server

Options not passed to the constructor are taken from the global settings
in =lib/LocalSite.cfg=.

=cut

sub new {
  my $class = shift;
  my $session = shift;

  # setting the SESSION var rather early; some foswiki engines do that a bit
  # later than calling for a login manager. that will cause all sorts of
  # problems as the Foswiki::Func API can't be used then. see Item10084
  $Foswiki::Plugins::SESSION = $session;

  my $this = {
    ldap => undef,    # connect later
    error => undef,
    host => $Foswiki::cfg{ShibLdap}{Host} || 'localhost',
    base => $Foswiki::cfg{ShibLdap}{Base} || '',
    port => $Foswiki::cfg{ShibLdap}{Port} || 389,
    version => $Foswiki::cfg{ShibLdap}{Version} || 3,

    userBase => $Foswiki::cfg{ShibLdap}{UserBase}
      || [$Foswiki::cfg{ShibLdap}{Base}]
      || [],

    userScope => $Foswiki::cfg{ShibLdap}{UserScope}
      || 'sub',

    groupBase => $Foswiki::cfg{ShibLdap}{GroupBase}
      || $Foswiki::cfg{ShibLdap}{Base}
      || [],

    groupScope => $Foswiki::cfg{ShibLdap}{GroupScope}
      || 'sub',

    # Hash keys that will be used as the just-in-memory 'cache' for the 
    # Shib-delivered attributes (loginName, wikiName, emails, groups).
    shibUserLoginName => '',
    shibUserWikiName => '',
    shibUserEmailAddresses => '',
    shibUserEmailsForLoginName => {}, 
    shibUserGroupNames => {},
	shibUserAllGroups => '',
    
    # Shibboleth-specific configuration options, along with re-used Ldap config options
    # that are still applicable.
    shibLoginEnvironmentVariable => $Foswiki::cfg{ShibLdap}{ShibLoginEnvironmentVariable} || 'REMOTE_USER',
    shibMailAttribute => $Foswiki::cfg{ShibLdap}{ShibMailAttribute} || 'mail',
    shibWikiNameAttribute => $Foswiki::cfg{ShibLdap}{ShibWikiNameAttributes} || 'eduPersonPrincipalName',

    normalizeWikiName => $Foswiki::cfg{ShibLdap}{NormalizeWikiNames},
    normalizeLoginName => $Foswiki::cfg{ShibLdap}{NormalizeLoginNames},
    caseSensitiveLogin => $Foswiki::cfg{ShibLdap}{CaseSensitiveLogin} || 0,
    ldapNormalizeGroupName => $Foswiki::cfg{ShibLdap}{LdapNormalizeGroupNames},
    shibNormalizeGroupName => $Foswiki::cfg{ShibLdap}{ShibNormalizeGroupNames},

    shibGroupAttribute => $Foswiki::cfg{ShibLdap}{ShibGroupAttribute} || 'isMemberOf',
    shibMapGroups => $Foswiki::cfg{ShibLdap}{ShibMapGroups} || 1,
	shibRewriteGroupNames => $Foswiki::cfg{ShibLdap}{ShibRewriteGroupNames} || {},
    shibMergeWithLdapGroups => $Foswiki::cfg{ShibLdap}{ShibMergeWithLdapGroups} || 1,		
	
	wikiNameAttribute => $Foswiki::cfg{ShibLdap}{WikiNameAttributes}
      || $Foswiki::cfg{ShibLdap}{WikiNameAttribute}
      || 'cn',

    wikiNameAliases => $Foswiki::cfg{ShibLdap}{WikiNameAliases} || '',	    

    groupAttribute => $Foswiki::cfg{ShibLdap}{GroupAttribute} || 'description',
    primaryGroupAttribute => $Foswiki::cfg{ShibLdap}{PrimaryGroupAttribute} || 'gidNumber',
    groupFilter => $Foswiki::cfg{ShibLdap}{GroupFilter} || 'objectClass=groupOfNames',
    memberAttribute => $Foswiki::cfg{ShibLdap}{MemberAttribute} || 'hasMember',
    innerGroupAttribute => $Foswiki::cfg{ShibLdap}{InnerGroupAttribute} || '',
    memberIndirection => $Foswiki::cfg{ShibLdap}{MemberIndirection} || 0,
    nativeGroupsBackoff => $Foswiki::cfg{ShibLdap}{WikiGroupsBackoff} || 0,
    bindDN => $Foswiki::cfg{ShibLdap}{BindDN} || '',
    bindPassword => $Foswiki::cfg{ShibLdap}{BindPassword} || '',
    mapGroups => $Foswiki::cfg{ShibLdap}{MapGroups} || 0,
    rewriteGroups => $Foswiki::cfg{ShibLdap}{RewriteGroups} || {},
    rewriteWikiNames => $Foswiki::cfg{ShibLdap}{RewriteWikiNames} || {},
    mergeGroups => $Foswiki::cfg{ShibLdap}{MergeGroups} || 0,

    # These will not likely be actually used, but left in just in case
    # someone comes up with a use.
    loginAttribute => $Foswiki::cfg{ShibLdap}{LoginAttribute} || 'uid',
    allowChangePassword => $Foswiki::cfg{ShibLdap}{AllowChangePassword} || 0,
    loginFilter => $Foswiki::cfg{ShibLdap}{LoginFilter} || 'objectClass=posixAccount',
    mailAttribute => $Foswiki::cfg{ShibLdap}{MailAttribute} || 'mail',

    exclude => $Foswiki::cfg{ShibLdap}{Exclude}
      || 'WikiGuest, ProjectContributor, RegistrationAgent, UnknownUser, AdminGroup, NobodyGroup, AdminUser, admin, guest',

    pageSize => $Foswiki::cfg{ShibLdap}{PageSize},
    isConnected => 0,
    maxCacheAge => $Foswiki::cfg{ShibLdap}{MaxCacheAge},
    preCache => $Foswiki::cfg{ShibLdap}{Precache},

    useSASL => $Foswiki::cfg{ShibLdap}{UseSASL} || 0,
    saslMechanism => $Foswiki::cfg{ShibLdap}{SASLMechanism} || 'PLAIN CRAM-MD4 EXTERNAL ANONYMOUS',

    useTLS => $Foswiki::cfg{ShibLdap}{UseTLS} || 0,
    tlsVerify => $Foswiki::cfg{ShibLdap}{TLSVerify} || 'require',
    tlsSSLVersion => $Foswiki::cfg{ShibLdap}{TLSSSLVersion} || 'tlsv1',
    tlsCAFile => $Foswiki::cfg{ShibLdap}{TLSCAFile} || '',
    tlsCAPath => $Foswiki::cfg{ShibLdap}{TLSCAPath} || '',
    tlsClientCert => $Foswiki::cfg{ShibLdap}{TLSClientCert} || '',
    tlsClientKey => $Foswiki::cfg{ShibLdap}{TLSClientKey} || '',

    secondaryPasswordManager => $Foswiki::cfg{ShibLdap}{SecondaryPasswordManager} || '',
    @_
  };
  bless($this, $class);

  $this->{session} = $session;

  $this->{preCache} = 1 unless defined $this->{preCache};
  $this->{pageSize} = 200 unless defined $this->{pageSize};

  unless (ref($this->{userBase})) {
    push @{$this->{userBase}}, $this->{userBase};
  }

  unless (ref($this->{groupBase})) {
    push @{$this->{groupBase}}, $this->{groupBase};
  }

  if ($this->{useSASL}) {
    #writeDebug("will use SASL authentication");
    require Authen::SASL;
  }

  # protect against accidental misconfiguration, that might lead
  # to an infinite loop during authorization etc.
  if ($this->{secondaryPasswordManager} eq 'Foswiki::Users::LdapPasswdUser') {
    writeWarning("hey, you want infinite loops? naw.");
    $this->{secondaryPasswordManager} = '';
  }

  if ($this->{secondaryPasswordManager} eq 'none') {
    $this->{secondaryPasswordManager} = '';
  }

  my $workArea = $session->{store}->getWorkArea('ShibLdapContrib');
  mkdir $workArea unless -d $workArea;
  $this->{cacheFile} = $workArea . '/cache.db';

  # normalize normalization flags
  $this->{normalizeWikiName} = $Foswiki::cfg{ShibLdap}{NormalizeWikiName}
    unless defined $this->{normalizeWikiName};
  $this->{normalizeWikiName} = 1
    unless defined $this->{normalizeWikiName};
  $this->{normalizeLoginName} = $Foswiki::cfg{ShibLdap}{NormalizeLoginName}
    unless defined $this->{normalizeLoginName};
  $this->{normalizeGroupName} = $Foswiki::cfg{ShibLdap}{NormalizeGroupName}
    unless defined $this->{normalizeGroupName};
  $this->{shibNormalizeGroupName} = $Foswiki::cfg{ShibLdap}{ShibNormalizeGroupName}
    unless defined $this->{shibNormalizeGroupName};

  @{$this->{wikiNameAttributes}} = split(/\s*,\s*/, $this->{wikiNameAttribute});
  @{$this->{shibWikiNameAttributes}} = split(/\s*,\s*/, $this->{shibWikiNameAttribute});

  # create exclude map
  my %excludeMap;
  if ($this->{caseSensitiveLogin}) {
    %excludeMap = map { $_ => 1 } split(/\s*,\s*/, $this->{exclude});
  } else {
    %excludeMap = map { $_ => 1, lc($_) => 1 } split(/\s*,\s*/, $this->{exclude});
  }
  $this->{excludeMap} = \%excludeMap;

  # creating alias map
  my %aliasMap = ();
  foreach my $alias (split(/\s*,\s*/, $this->{wikiNameAliases})) {
    if ($alias =~ /^\s*(.+?)\s*=\s*(.+?)\s*$/) {
      $aliasMap{$1} = $2;
    }
  }
  $this->{wikiNameAliases} = \%aliasMap;

  # default value for cache expiration is every 24h
  $this->{maxCacheAge} = 86400 unless defined $this->{maxCacheAge};

  writeDebug("constructed a new ShibLdapContrib object");

  return $this;
}

=pod

---++ getShibLdapContrib($session) -> $ldap

Returns a standard singleton Foswiki::Contrib::ShibLdapContrib object based on the site-wide
configuration. 

=cut

sub getShibLdapContrib {
  my $session = shift;

  my $obj = $sharedShibLdapContrib{$session};
  return $obj if $obj;

  $obj = new Foswiki::Contrib::ShibLdapContrib($session);
  $obj->initCache();
  $sharedShibLdapContrib{$session} = $obj;

  return $obj;
}

=pod

---++ connect($dn, $passwd, $host, $port) -> $boolean

Connect to LDAP server. If a $dn parameter and a $passwd is given then a bind is done.
Otherwise the communication is anonymous. You don't have to connect() explicitely
by calling this method. The methods below will do that automatically when needed.

=cut

sub connect {
  my ($this, $dn, $passwd, $host, $port) = @_;

  writeDebug("called connect");
  #writeDebug("dn=$dn", 2) if $dn;
  #writeDebug("passwd=***", 2) if $passwd;

  $host ||= $this->{host};
  $port ||= $this->{port};

  $this->{ShibLdap} = Net::LDAP->new(
    $host,
    port => $port,
    version => $this->{version},
  );
  unless ($this->{ShibLdap}) {
    $this->{error} = "failed to connect to $this->{host}";
    $this->{error} .= ": $@" if $@;
    return 0;
  }

  # TLS bind
  if ($this->{useTLS}) {
    #writeDebug("using TLS");
    my %args = (
      verify => $this->{tlsVerify},
      cafile => $this->{tlsCAFile},
      capath => $this->{tlsCAPath},
    );
    $args{"clientcert"} = $this->{tlsClientCert} if $this->{tlsClientCert};
    $args{"clientkey"} = $this->{tlsClientKey} if $this->{tlsClientKey};
    $args{"sslversion"} = $this->{tlsSSLVersion} if $this->{tlsSSLVersion};
    my $msg = $this->{ShibLdap}->start_tls(%args);
    writeWarning($msg->{errorMessage}) if exists $msg->{errorMessage};
  }

  $passwd = $this->fromSiteCharSet($passwd) if $passwd;

  # authenticated bind
  my $msg;
  if (defined($dn)) {
    die "illegal call to connect()" unless defined($passwd);
    $msg = $this->{ShibLdap}->bind($dn, password => $passwd);
    #writeDebug("bind for $dn");
  }

  # proxy user
  elsif ($this->{bindDN} && $this->{bindPassword}) {

    if ($this->{useSASL}) {
      # sasl bind
      my $sasl = Authen::SASL->new(
        mechanism => $this->{saslMechanism},    #'DIGEST-MD5 PLAIN CRAM-MD5 EXTERNAL ANONYMOUS',
        callback => {
          user => $this->{bindDN},
          pass => $this->{bindPassword},
        },
      );
      #writeDebug("sasl bind to $this->{bindDN}");
      $msg = $this->{ShibLdap}->bind($this->{bindDN}, sasl => $sasl, version => $this->{version});
    } else {
      # simple bind
      #writeDebug("proxy bind");
      $msg = $this->{ShibLdap}->bind($this->{bindDN}, password => $this->{bindPassword});
    }
  }

  # anonymous bind
  else {
    writeDebug("anonymous bind");
    $msg = $this->{ShibLdap}->bind;
  }

  $this->{isConnected} = ($this->checkError($msg) == LDAP_SUCCESS) ? 1 : 0;
  writeDebug("failed to bind") unless $this->{isConnected};
  return $this->{isConnected};
}

=pod

---++ disconnect()

Unbind the LDAP object from the server. This method can be used to force
a reconnect and possibly rebind as a different user.

=cut

sub disconnect {
  my $this = shift;

  return unless defined($this->{ShibLdap}) && $this->{isConnected};

  #writeDebug("called disconnect()");
  $this->{ShibLdap}->unbind();
  $this->{ShibLdap} = undef;
  $this->{isConnected} = 0;
}

=pod

---++ finish()

finalize this ldap object.

=cut

sub finish {
  my $this = shift;

  return if $this->{isFinished};
  $this->{isFinished} = 1;

  #writeDebug("finishing");

  $this->disconnect();
  delete $sharedShibLdapContrib{$this->{session}};

  undef $this->{cacheDB};
  untie %{$this->{data}};
}

=pod

---++ checkError($msg) -> $errorCode

Private method to check a Net::LDAP::Message object for an error, sets
$ldap->{error} and returns the ldap error code. This method is called
internally whenever a message object is returned by the server. Use
$ldap->getError() to return the actual error message.

=cut

sub checkError {
  my ($this, $msg) = @_;

  my $code = $this->{code} = $msg->code();
  if ($code == LDAP_SUCCESS || $code == LDAP_REFERRAL) {
    $this->{error} = undef;
  } else {
    $this->{error} = $code . ': ' . $msg->error();
    #writeDebug($this->{error});
  }

  return $code;
}

=pod

---++ getError() -> $errorMsg

Returns the error message of the last LDAP action or undef it no
error occured.

=cut

sub getError {
  return $_[0]->{error};
}

=pod

---++ getCode() -> $errorCode

Returns the error code of the last LDAP action

=cut

sub getCode {
  return $_[0]->{code};
}

=pod

---++ getAccount($loginName) -> Net::LDAP::Entry object

Fetches an account entry from the database and returns a Net::LDAP::Entry
object on success and undef otherwise. Note, the login name is match against
the attribute defined in $ldap->{loginAttribute}. Account records are 
search using $ldap->{loginFilter} in one of the the subtrees defined in
$ldap->{userBase}.

=cut

sub getAccount {
  my ($this, $loginName) = @_;

  return undef if $this->{excludeMap}{$loginName};
  #writeDebug("called getAccount($loginName)");

  # take care of login case
  $loginName = lc($loginName) unless $this->{caseSensitiveLogin};

  my $loginFilter = $this->{loginFilter};
  $loginFilter = "($loginFilter)" unless $loginFilter =~ /^\(.*\)$/;
  my $filter = '(&' . $loginFilter . '(' . $this->{loginAttribute} . '=' . $loginName . '))';

  my $entry;
  foreach my $userBase (@{$this->{userBase}}) {
    my $msg = $this->search(
      filter => $filter,
      base => $userBase,
      deref => "always",
      callback => sub {
        my (undef, $result) = @_;
        return unless defined $result;
        $entry = $result;
      }
    );

    return $entry if $entry;
  }

  $this->{error} = 'Login invalid';
  #writeDebug($this->{error});

  return;
}

=pod

---++ search($filter, %args) -> $msg

Returns an Net::LDAP::Search object for the given query on success and undef
otherwise. If $args{base} is not defined $ldap->{base} is used.  If $args{scope} is not
defined 'sub' is used (searching down the subtree under $args{base}. If no $args{sizelimit} is
set all matching records are returned.  The $attrs is a reference to an array
of all those attributes that matching entries should contain.  If no $args{attrs} is
defined all attributes are returned.

If undef is returned as an error occured use $ldap->getError() to get the
cleartext message of this search() operation.

Typical usage:
<verbatim>
$ldap->search(
  filter=>'uid=TestUser',
  callback => sub {
    my ($ldap, $entry) = @_;
    return unless defined $entry;
    # process entry
  }
);
</verbatim>

=cut

sub search {
  my ($this, %args) = @_;

  $args{base} = $this->{base} unless $args{base};
  $args{base} = $this->fromSiteCharSet($args{base});
  $args{scope} = 'sub' unless $args{scope};
  $args{sizelimit} = 0 unless $args{sizelimit};
  $args{attrs} = ['*'] unless $args{attrs};
  $args{filter} = $this->fromSiteCharSet($args{filter}) if $args{filter};

  if (defined($args{callback}) && !defined($args{_origCallback})) {

    $args{_origCallback} = $args{callback};

    $args{callback} = sub {
      my ($ldap, $entry) = @_;

      # bail out when done
      return unless defined $entry;

      # follow references
      if ($entry->isa("Net::LDAP::Reference")) {
        foreach my $link ($entry->references) {
          #writeDebug("following reference $link");
          $this->_followLink($link, %args);
        }
      } else {
        # call the orig callback
        $args{_origCallback}($ldap, $entry);
      }
    };
  }

  if ($Foswiki::cfg{ShibLdap}{Debug}) {
    my $attrString = join(',', @{$args{attrs}});
    #writeDebug("called search(filter=$args{filter}, base=$args{base}, scope=$args{scope}, sizelimit=$args{sizelimit}, attrs=$attrString)");
  }

  unless ($this->{ShibLdap}) {
    unless ($this->connect()) {
      #writeDebug("error in search: " . $this->getError());
      return undef;
    }
  }

  my $msg = $this->{ShibLdap}->search(%args);
  my $errorCode = $this->checkError($msg);

  # we set a sizelimit so it is ok that it exceeds
  if ($args{sizelimit} && $errorCode == LDAP_SIZELIMIT_EXCEEDED) {
    #writeDebug("sizelimit exceeded");
    return $msg;
  }

  if ($errorCode == LDAP_REFERRAL) {
    unless ($this->{_followingLink}) {
      my @referrals = $msg->referrals;
      foreach my $link (@referrals) {
        #writeDebug("following referral $link");
        $this->_followLink($link, %args);
      }
    }
  } elsif ($errorCode != LDAP_SUCCESS) {
    #writeDebug("error in search: " . $this->getError());
    return undef;
  }

  return $msg;
}

sub _followLink {
  my ($this, $link, %args) = @_;

  #writeDebug("following ldap url $link");
  #return if $this->{_followingLink};
  $this->{_followingLink} = 1;

  my $uri = URI::ldap->new($link);

  # SMELL: cache multiple ldap connections

  # remember old connection
  my $oldLdap = $this->{ShibLdap};
  my %thisArgs = %args;
  $thisArgs{base} = $uri->dn;
  $thisArgs{port} = $uri->port;

  # trick in new connection
  $this->connect(undef, undef, $uri->host, $uri->port);
  $this->search(%thisArgs);
  $this->disconnect;

  # restore old connection
  $this->{ShibLdap} = $oldLdap;
  $this->{isConnected} = 1 if defined $oldLdap;

  $this->{_followingLink} = 0;

  #writeDebug("done following ldap url $link");
}

=pod

---++ cacheBlob($entry, $attribute, $refresh) -> $pubUrlPath

Takes an Net::LDAP::Entry and an $attribute name, and stores its value into a
file. Returns the pubUrlPath to it. This can be used to store binary large
objects like images (jpegPhotos) into the filesystem accessible to the httpd
which can serve it in return to the client browser. 

Filenames containing the blobs are named using a hash value that is generated
using its DN and the actual attribute name whose value is extracted from the 
database. If the blob already exists in the cache it is _not_ extracted once
again except the $refresh parameter is defined.

Typical usage:
<verbatim>
my $blobUrlPath = $ldap->cacheBlob($entry, $attr);
</verbatim>

=cut

sub cacheBlob {
  my ($this, $entry, $attr, $refresh) = @_;

  #writeDebug("called cacheBlob()");
  require Digest::MD5;

  my $systemWeb = $Foswiki::cfg{SystemWebName};
  my $dir = $Foswiki::cfg{PubDir} . '/' . $systemWeb . '/ShibLdapContrib';
  mkdir $dir unless -d $dir;
  $dir .= '/blobs';
  mkdir $dir unless -d $dir;

  my $key = Digest::MD5::md5_hex($entry->dn() . $attr);
  my $fileName = $dir . '/' . $key;

  if ($refresh || !-f $fileName) {
    #writeDebug("caching blob for $attr to $fileName");
    my $value = $entry->get_value($attr);
    return undef unless defined $value;
    mkdir($dir, 0775) unless -e $dir;

    open(FILE, ">$fileName") || die "can't open $fileName";
    binmode(FILE);
    print FILE $value;
    close(FILE);
  } else {
    #writeDebug("already got blob");
  }

  #writeDebug("done cacheBlob()");
  return $Foswiki::cfg{PubUrlPath} . '/' . $systemWeb . '/ShibLdapContrib/blobs/' . $key;
}

=pod

---++ initCache()

loads/connects to the LDAP cache

=cut

sub initCache {
  my $this = shift;

  return
    unless $Foswiki::cfg{UserMappingManager} =~ /ShibLdapUserMapping/
      || $Foswiki::cfg{PasswordManager} =~ /LdapPasswdUser/;

  writeDebug("called initCache");

  # open cache
  writeDebug("opening ldap cache from $this->{cacheFile}");
  $this->{cacheDB} = tie %{$this->{data}}, 'DB_File', $this->{cacheFile}, O_CREAT | O_RDWR, 0664, $DB_HASH
    or die "Cannot open file $this->{cacheFile}: $!";

  # refresh by user interaction
  my $refresh = '';
  $refresh = CGI::param('refreshldap') || '';
  $refresh = 1 if $refresh eq 'on';
  $refresh = 2 if $refresh eq 'force';
  $refresh = 0 unless $refresh;

  if ($this->{maxCacheAge} > 0) {    # is cache expiration enabled

    # compute age of data
    my $cacheAge = 9999999999;
    my $now = time();
    my $lastUpdate = $this->{data}{lastUpdate} || 0;
    $cacheAge = $now - $lastUpdate if $lastUpdate;

    # don't refresh within 60 seconds
    if ($cacheAge < 10) {
      $refresh = 0;
      writeDebug("suppressing cache refresh within 10 seconds");
    } else {
      $refresh = 1 if $cacheAge > $this->{maxCacheAge};
    }

    writeDebug("cacheAge=$cacheAge, maxCacheAge=$this->{maxCacheAge}, lastUpdate=$lastUpdate, refresh=$refresh");
  }

  # clear to reload it
  $this->refreshCache($refresh);
}

=pod

---++ refreshCache($mode) -> $boolean

download all relevant records from the LDAP server and
store it into a database.

   * mode = 0: no refresh
   * mode = 1: normal refersh
   * mode = 2: nuke previous decisions on wikiName clashes

=cut

sub refreshCache {
  my ($this, $mode) = @_;

  return unless $mode;

  writeDebug("called refreshCache(mode=$mode)");

  $this->{_refreshMode} = $mode;

  # create a temporary tie
  my $tempCacheFile = $this->{cacheFile} . '_tmp';
  if (-e $tempCacheFile) {
    writeDebug("already refreshing cache");
    return 0;
  }

  my %tempData;
  my $tempCache = tie %tempData, 'DB_File', $tempCacheFile, O_CREAT | O_RDWR, 0664, $DB_HASH
    or die "Cannot open file $tempCacheFile: $!";

  # precache the LDAP directory if enabled in configuration file
  if ($this->{preCache}) {
    writeDebug("precaching is ON.");

    my $isOk = 1;

    foreach my $userBase (@{$this->{userBase}}) {
      $isOk = $this->refreshUsersCache(\%tempData, $userBase);
      last unless $isOk;
    }

    if ($isOk && $this->{mapGroups}) {
      foreach my $groupBase (@{$this->{groupBase}}) {
        $isOk = $this->refreshGroupsCache(\%tempData, $groupBase);
        writeDebug("groupbase: $groupBase, isOk: '$isOk'");
        last unless $isOk;
      }
    }

    unless ($isOk) {    # we had an error: keep the old cache til the error is resolved
      writeDebug("error in rebuilding cache");
      undef $tempCache;
      untie %tempData;
      unlink $tempCacheFile;
      return 0;
    }
  }

  writeDebug("flushing db to disk");
  $tempData{lastUpdate} = time();
  $tempCache->sync();
  undef $tempCache;
  untie %tempData;

  # try to be transactional
  undef $this->{cacheDB};
  untie %{$this->{data}};

  writeDebug("replacing working copy");
  rename $tempCacheFile, $this->{cacheFile};

  # reconnect hash
  $this->{cacheDB} = tie %{$this->{data}}, 'DB_File', $this->{cacheFile}, O_CREAT | O_RDWR, 0664, $DB_HASH
    or die "Cannot open file $this->{cacheFile}: $!";

  undef $this->{_refreshMode};

  return 1;
}

=pod

---++ refreshUsersCache($data, $userBase) -> $boolean

download all user records from the LDAP server and cache it into the
given hash reference

returns true if new records have been loaded

=cut

sub refreshUsersCache {
  my ($this, $data, $userBase) = @_;

  writeDebug("called refreshUsersCache($userBase)");
  $data ||= $this->{data};
  $userBase ||= $this->{base};

  # prepare search
  my %args = (
    filter => $this->{loginFilter},
    base => $userBase,
    scope => $this->{userScope},
    deref => "always",
    attrs => [$this->{loginAttribute}, $this->{mailAttribute}, $this->{primaryGroupAttribute}, @{$this->{wikiNameAttributes}}],
  );

  # use the control LDAP extension only if a valid pageSize value has been provided
  my $page;
  my $cookie;
  if ($this->{pageSize} > 0) {
    require Net::LDAP::Control::Paged;
    $page = Net::LDAP::Control::Paged->new(size => $this->{pageSize});
    $args{control} = [$page];
    writeDebug("reading users from cache with page size=$this->{pageSize}");
  } else {
    writeDebug("reading users from cache in one chunk");
  }

  # read pages
  my $nrRecords = 0;
  my %wikiNames = ();
  my %loginNames = ();
  my $gotError = 0;

  if (defined $data->{WIKINAMES}) {
    %wikiNames = map { $_ => 1 } $data->{WIKINAMES};
  }
  if (defined $data->{LOGINNAMES}) {
    %loginNames = map { $_ => 1 } $data->{LOGINNAMES};
  }

  $args{callback} = sub {
    my ($ldap, $entry) = @_;
    $this->cacheUserFromEntry($entry, $data, \%wikiNames, \%loginNames) && $nrRecords++;
  };

  while (1) {

    # perform search
    my $mesg = $this->search(%args);
    unless ($mesg) {
      writeWarning("error refreshing the user cache: " . $this->getError());
      $gotError = 1 unless $this->getCode() == LDAP_SIZELIMIT_EXCEEDED;    # continue on sizelimit exceeded
      last;
    }

    # only use cookies and pages if we are using this extension
    if ($page) {
      # get cookie from paged control to remember the offset
      my ($resp) = $mesg->control(LDAP_CONTROL_PAGED) or last;

      $cookie = $resp->cookie or last;
      if ($cookie) {
        # set cookie in paged control
        $page->cookie($cookie);
      } else {
        # found all
        #writeDebug("ok, no more cookie");
        last;
      }
    } else {
      # one chunk ends here
      last;
    }
  }    # end reading pages

  writeDebug("done reading person pages");

  # clean up
  if ($cookie) {
    $page->cookie($cookie);
    $page->size(0);
    $this->search(%args);
  }

  # check for error
  return 0 if $gotError;

  # resolving WikiName clashes
  $nrRecords += $this->resolveWikiNameClashes($data, \%wikiNames, \%loginNames);

  # remember list of all user names
  $data->{WIKINAMES} = join(',', keys %wikiNames);
  $data->{LOGINNAMES} = join(',', keys %loginNames);

  writeDebug("got $nrRecords person keys in cache");

  return 1;
}

=pod

---++ resolveWikiNameClashes($data, %wikiNames, %loginNames) -> $integer

if there have been name clashes during cacheUserFromEntry() those entry records
have not yet been added to the cache. They are kept until all clashes have been
found and a deterministic renaming scheme can be applied. Clashed WikiNames will be
enumerated - !WikiName1, !WikiName2, !WikiName3 etc - and finally added to the database.
The renaming is kept stable by sorting the dn entry of all clashed entries.

returns the number of additional entries that have been cached

=cut

sub resolveWikiNameClashes {
  my ($this, $data, $wikiNames, $loginNames) = @_;

  return 0 unless $this->{_wikiNameClaches};

  $data ||= $this->{data};

  my %suffixes = ();
  my $nrRecords = 0;
  my $refreshMode = $this->{_refreshMode} || 0;
  foreach my $item (sort { $a->{loginName} cmp $b->{loginName} } values %{$this->{_wikiNameClaches}}) {

    my $wikiName = $item->{wikiName};
    my $loginName = $item->{loginName};
    my $dn = $item->{dn};
    my $prevWikiName = ($refreshMode < 2) ? $this->getWikiNameOfDn($dn) : '';
    my $newWikiName;
    my $prevDN;

    if ($prevWikiName) {
      writeDebug("found prevWikiName=$prevWikiName for $dn");
      $newWikiName = $prevWikiName;
    } else {

      # search for a new wikiname
      do {
        $suffixes{$wikiName}++;
        $newWikiName = $wikiName . $suffixes{$wikiName};
        $prevDN = $this->getDnOfWikiName($newWikiName);
      } until (!$prevDN || $prevDN eq $dn);
    }

    writeDebug("processing clash of loginName=$loginName on wikiName=$wikiName, dn=$item->{dn}, resolves to newWikiName=$newWikiName");

    $this->cacheUserFromEntry($item->{entry}, $data, $wikiNames, $loginNames, $newWikiName);
    $nrRecords++;
  }

  delete $this->{_wikiNameClaches};

  return $nrRecords;
}

=pod

---++ refreshGroupsCache($data, $groupBase) -> $boolean

download all group records from the LDAP server

returns true if new records have been loaded

=cut

sub refreshGroupsCache {
  my ($this, $data, $groupBase) = @_;

  $data ||= $this->{data};

  writeDebug("called refreshGroupsCache($groupBase)");
  # prepare search
  my %args = (
    filter => $this->{groupFilter},
    base => $groupBase,
    scope => $this->{groupScope},
    deref => "always",
    attrs => [$this->{groupAttribute}, $this->{memberAttribute}, $this->{innerGroupAttribute}, $this->{primaryGroupAttribute}],
  );

  # use the control LDAP extension only if a valid pageSize value has been provided
  my $page;
  my $cookie;
  if ($this->{pageSize} > 0) {
    require Net::LDAP::Control::Paged;
    $page = Net::LDAP::Control::Paged->new(size => $this->{pageSize});
    $args{control} = [$page];
  } else {
    #writeDebug("reading group from cache in one chunk");
  }

  # read pages
  my $nrRecords = 0;
  my %groupNames = map { $_ => 1 } @{$this->getGroupNames($data)};
  my $gotError = 0;

  $args{callback} = sub {
    my ($ldap, $entry) = @_;
    $this->cacheGroupFromEntry($entry, $data, \%groupNames) && $nrRecords++;
  };

  while (1) {

    # perform search
    my $mesg = $this->search(%args);
    unless ($mesg) {
      #writeDebug("oops, no result querying for groups");
      writeWarning("error refeshing the groups cache: " . $this->getError());
      last;
    }

    # only use cookies and pages if we are using this extension
    if ($page) {
      # get cookie from paged control to remember the offset
      my ($resp) = $mesg->control(LDAP_CONTROL_PAGED) or last;

      $cookie = $resp->cookie or last;
      if ($cookie) {
        # set cookie in paged control
        $page->cookie($cookie);
      } else {
        # found all
        #writeDebug("ok, no more cookie");
        last;
      }
    } else {
      last;
    }
  }    # end reading pages

  writeDebug("done reading group pages");

  # clean up
  if ($cookie) {
    $page->cookie($cookie);
    $page->size(0);
    $this->search(%args);
  }

  # check for error
  return 0 if $gotError;

  # check for primary group membership
  if ($this->{_primaryGroup}) {
    foreach my $groupId (keys %{$this->{_primaryGroup}}) {
      my $groupName = $this->{_groupId}{$groupId};
      next unless $groupName;
      foreach my $member (keys %{$this->{_primaryGroup}{$groupId}}) {
        #writeDebug("adding $member to its primary group $groupName");
        $this->{_groups}{$groupName}{$member} = 1;
      }
    }
  }

  # assert group members to data store
  foreach my $groupName (keys %{$this->{_groups}}) {

    #writeDebug("adding group $groupName");
    my %members = ();
    foreach my $member (keys %{$this->{_groups}{$groupName}}) {

      # groups may store DNs to members instead of a memberUid, in this case we
      # have to lookup the corresponding loginAttribute
      if ($this->{memberIndirection}) {
        #writeDebug("following indirection for $member");
        my $memberName = $data->{"DN2U::$member"};
        if ($memberName) {
          $members{$memberName} = 1;
        } else {
          #writeWarning("oops, $member not found, but member of $groupName");
        }
      } else {
        $members{$member} = 1;
      }
    }

    $data->{"GROUPS::$groupName"} = join(',', sort keys %members);
    undef $this->{_groups}{$groupName};
  }
  undef $this->{_groups};

  # remember list of all groups
  $data->{GROUPS} = join(',', sort keys %groupNames);

  writeDebug("got $nrRecords group keys in cache");

  return 1;
}

=pod

---++ cacheUserFromEntry($entry, $data, $wikiNames, $loginNames, $wikiName) -> $boolean

store a user LDAP::Entry to our internal cache 

If the $wikiName parameter is given explicitly then this will be the name under which this record
will be cached.

returns true if new records have been created

=cut

sub cacheUserFromEntry {
  my ($this, $entry, $data, $wikiNames, $loginNames, $wikiName) = @_;

  #writeDebug("called cacheUserFromEntry()");

  $data ||= $this->{data};
  $wikiNames ||= {};
  $loginNames ||= {};

  my $dn = $entry->dn();

  # 1. get it
  my $loginName = $entry->get_value($this->{loginAttribute});
  $loginName =~ s/^\s+//o;
  $loginName =~ s/\s+$//o;
  unless ($loginName) {
    #writeDebug("no loginName for $dn ... skipping");
    return 0;
  }
  $loginName = $this->fromLdapCharSet($loginName);

  # 2. normalize
  $loginName = lc($loginName) unless $this->{caseSensitiveLogin};
  $loginName = $this->normalizeLoginName($loginName) if $this->{normalizeLoginName};
  return 0 if $this->{excludeMap}{$loginName};

  # construct the wikiName
  my $isExplicitWikiName = (defined $wikiName) ? 1 : 0;
  if ($isExplicitWikiName) {
    #writeDebug("found explicit wikiName '$wikiName' for $dn");
  } else {

    # 0. get previously used wikiName
    my $refreshMode = $this->{_refreshMode} || 0;
    my $prevWikiName = ($refreshMode < 2) ? $this->getWikiNameOfDn($dn) : '';

    # keep a wikiName once it has been computed
    if ($prevWikiName) {
      #writeDebug("found prevWikiName=$prevWikiName for $dn");
      $wikiName = $prevWikiName;
    } else {

      # 1. compute a new wikiName
      my @wikiName = ();
      foreach my $attr (@{$this->{wikiNameAttributes}}) {
        my $value = $entry->get_value($attr);
        next unless $value;
        $value =~ s/^\s+//o;
        $value =~ s/\s+$//o;
        $value = $this->fromLdapCharSet($value);
        #writeDebug("$attr=$value");
        push @wikiName, $value;
      }
      $wikiName = join(" ", @wikiName);

      unless ($wikiName) {
        #$wikiName = $loginName;
        #writeWarning("no WikiNameAttributes found for $dn ... deriving WikiName from LoginName: '$wikiName'");
        writeWarning("no WikiNameAttributes found for $dn ... ignoring");
        return 0;
      }

      # 2. rewrite
      my $oldWikiName = $wikiName;
      foreach my $pattern (keys %{$this->{rewriteWikiNames}}) {
        my $subst = $this->{rewriteWikiNames}{$pattern};
        if ($wikiName =~ /^(?:$pattern)$/) {
          my $arg1 = $1;
          my $arg2 = $2;
          my $arg3 = $3;
          my $arg4 = $4;
          my $arg5 = $5;
          $arg1 = '' unless defined $arg1;
          $arg2 = '' unless defined $arg2;
          $arg3 = '' unless defined $arg3;
          $arg4 = '' unless defined $arg4;
          $subst =~ s/\$1/$arg1/g;
          $subst =~ s/\$2/$arg2/g;
          $subst =~ s/\$3/$arg3/g;
          $subst =~ s/\$4/$arg4/g;
          $subst =~ s/\$5/$arg5/g;
          $wikiName = $subst;
          #writeDebug("rewriting '$oldWikiName' to '$wikiName' using rule $pattern");
          last;
        }
      }

      # 3. normalize
      if ($this->{normalizeWikiName}) {
        $wikiName = $this->normalizeWikiName($wikiName);
      }

      # 4. aliasing based on WikiName
      my $alias = $this->{wikiNameAliases}{$wikiName};
      if ($alias) {
        #writeDebug("using alias $alias for $wikiName");
        $wikiName = $alias;
      }

      # 5. check if this dn maps to a wikiName already in use by another dn before
      my $prevDN = $this->getDnOfWikiName($wikiName) || '';
      if ($prevDN && $prevDN ne $dn) {

        writeWarning("$dn clashes with wikiName $wikiName already in use by $prevDN ... renaming later");
        $this->{_wikiNameClaches}{$dn} = {
          entry => $entry,
          dn => $dn,
          wikiName => $wikiName,
          loginName => $loginName,
        };
        return 0;
      }

      # 6. check for name clashes within this transaction
      my $clashDN = $this->getDnOfWikiName($wikiName, $data);
      if (defined $clashDN) {
        if ($dn ne $clashDN) {
          writeWarning("$dn clashes with $clashDN on wikiName $wikiName ... renaming later");
          $this->{_wikiNameClaches}{$dn} = {
            entry => $entry,
            dn => $dn,
            wikiName => $wikiName,
            loginName => $loginName,
          };
        } else {
          # never reach: same dn found twice in same transaction
          writeWarning("$dn found twice in ldap search... ignoring second one");
        }
        return 0;
      }
    }
  }

  if (defined($loginNames->{$loginName})) {
    my $clashDN = $loginNames->{$loginName};
    if ($clashDN ne $dn) {
      if ($clashDN eq '1') {
        $clashDN = $data->{"U2DN::$loginName"} || '???';
      }
      writeWarning("$dn clashes with $clashDN on loginName $loginName ... please configure a unique loginName attribute");
      return 0;
    }
  }

  $wikiNames->{$wikiName} = $dn;
  $loginNames->{$loginName} = $dn;

  # get email addrs
  my $emails;
  @{$emails} = $entry->get_value($this->{mailAttribute});

  # get primary group
  if ($this->{primaryGroupAttribute}) {
    my $groupId = $entry->get_value($this->{primaryGroupAttribute});
    $this->{_primaryGroup}{$groupId}{$loginName} = 1 if $groupId;    # delayed
  }

  # store it
  #writeDebug("adding wikiName='$wikiName', loginName='$loginName', dn='$dn'");
  $data->{"U2W::$loginName"} = $wikiName;
  $data->{"W2U::$wikiName"} = $loginName;
  $data->{"DN2U::$dn"} = $loginName;
  $data->{"U2DN::$loginName"} = $dn;
  $data->{"U2EMAIL::$loginName"} = join(',', @$emails);

  if ($emails) {
    foreach my $email (@$emails) {
      $email =~ s/^\s+//o;
      $email =~ s/\s+$//o;
      my $prevMapping = $data->{"EMAIL2U::$email"};
      my %emails = ();
      if ($prevMapping) {
        %emails = map { $_ => 1 } split(/\s*,\s*/, $prevMapping);
      }
      $emails{$loginName} = $email;
      $data->{"EMAIL2U::$email"} = join(',', sort keys %emails);
    }
  }

  my %groupNames = map { $_ => 1 } @{$this->getGroupNames($data)};

  foreach my $groupName (keys %groupNames) {
    if (defined $data->{"::$groupName"}) {
      my $dnList = Foswiki::Sandbox::untaintUnchecked($data->{"GROUP2UNCACHEDMEMBERSDN::$groupName"}) || '';
      my @membersDn = split(/\s*;\s*/, $dnList);

    LOOP: {
        foreach my $memberDn (@membersDn) {
          if ($memberDn eq $dn) {

            #writeDebug("refreshing group $groupName to catch new members");
            removeGroupFromCache($this, $groupName, $data);
            checkCacheForGroupName($this, $groupName, $data);
            last LOOP;
          }
        }
      }
    }
  }
  return 1;
}

=pod

---++ cacheShibUserFromPerson($loginname, $data, $wikiName) -> $boolean

Store a Shib person to our "memory" cache, actually just entries in the ShibLdapContrib object {data}.
Note we don't need a DN (we'd have to make one up anyways), just the LoginName, wikiName, and emails
(and even the emails might not be that useful).

If the $wikiName parameter is given explicitly then this will be the wikiName used.

Returns true if new records have been created (always)

=cut

sub cacheShibUserFromPerson {
  my ($this, $loginName) = @_;

  writeDebug("called cacheShibUserFromPerson($loginName)");

  return 0 if $this->{excludeMap}{$loginName};

  my $wikiName; 

  # 1. compute a wikiName
  my @wikiName = ();
  #writeDebug("wikiname: @{$this->{shibWikiNameAttributes}}");
  foreach my $attr (@{$this->{shibWikiNameAttributes}}) {
     my $value = $ENV{$attr};
     #writeDebug("wikiname loop: attr: $attr, value: $value");
     next unless $value;
     $value =~ s/^\s+//o;
     $value =~ s/\s+$//o;
     $value = $this->fromLdapCharSet($value);
     #writeDebug("$attr=$value");
     push @wikiName, $value;
   }
   $wikiName = join(" ", @wikiName);
   $wikiName = Foswiki::Sandbox::untaintUnchecked($wikiName);
  
   unless ($wikiName) {
     $wikiName = $loginName;
     writeDebug("no WikiNameAttributes found for $loginName ... deriving WikiName from LoginName: '$wikiName'");
   }
  
   # 2. rewrite
   my $oldWikiName = $wikiName;
   foreach my $pattern (keys %{$this->{rewriteWikiNames}}) {
     my $subst = $this->{rewriteWikiNames}{$pattern};
     if ($wikiName =~ /^(?:$pattern)$/) {
       my $arg1 = $1;
       my $arg2 = $2;
       my $arg3 = $3;
       my $arg4 = $4;
       my $arg5 = $5;
       $arg1 = '' unless defined $arg1;
       $arg2 = '' unless defined $arg2;
       $arg3 = '' unless defined $arg3;
       $arg4 = '' unless defined $arg4;
       $subst =~ s/\$1/$arg1/g;
       $subst =~ s/\$2/$arg2/g;
       $subst =~ s/\$3/$arg3/g;
       $subst =~ s/\$4/$arg4/g;
       $subst =~ s/\$5/$arg5/g;
       $wikiName = $subst;
       writeDebug("rewriting '$oldWikiName' to '$wikiName' using rule $pattern");
       last;
     }
   }
   
   # 3. normalize
   if ($this->{normalizeWikiName}) {
     $wikiName = $this->normalizeWikiName($wikiName);
   }
  
   # 4. aliasing based on WikiName
   my $alias = $this->{wikiNameAliases}{$wikiName};
   if ($alias) {
     writeDebug("using alias $alias for $wikiName");
     $wikiName = $alias;
   }
  
   # 5. check if this wikiname maps to a login already in use by another wikiName before
   my %suffixes = ();
   my $prevLogin = $this->getLoginOfWikiName($wikiName) || '';
   if ($prevLogin && $prevLogin ne $loginName) {
		my $newWikiName;
			
	    # search for a new wikiname
	    do {
	        $suffixes{$wikiName}++;
	        $newWikiName = $wikiName . $suffixes{$wikiName};
	        $prevLogin = $this->getLoginOfWikiName($newWikiName);
	    } until (!$prevLogin || $prevLogin eq $loginName);
	    
	    writeDebug("processing clash of loginName=$loginName on wikiName=$wikiName, resolves to newWikiName=$newWikiName");
      }
   

  # get email addrs
  my @emails;
  @emails = split /;/, $ENV{$this->{shibMailAttribute}};
  @emails = Foswiki::Sandbox::untaintUnchecked(@emails);

  # store the Shib User information
  writeDebug("adding wikiName='$wikiName', loginName='$loginName', to our Shib cache entries");
  $this->{shibUserLoginName} = $loginName;
  $this->{shibUserWikiName}  = $wikiName;
  $this->{shibUserEmailAddresses}  = join(',', @emails);

  if (@emails) {
    foreach my $email (@emails) {
      $email =~ s/^\s+//o;
      $email =~ s/\s+$//o;
      $this->{shibUserEmailsForLoginName}{$email} = 1;
    }
  }

  return 1;
}

=pod

---++ cacheGroupFromEntry($entry, $data, $groupNames) -> $boolean

store a group LDAP::Entry to our internal cache 

returns true if new records have been created

=cut

sub cacheGroupFromEntry {
  my ($this, $entry, $data, $groupNames) = @_;

  $data ||= $this->{data};
  $groupNames ||= {};

  my $dn = $entry->dn();

  my $groupName = $entry->get_value($this->{groupAttribute});
  #writeDebug("caching group $groupName for $dn");
  unless ($groupName) {
    #writeDebug("no groupName for $dn ... skipping");
    return 0;
  }
  $groupName =~ s/^\s+//o;
  $groupName =~ s/\s+$//o;
  $groupName = $this->fromLdapCharSet($groupName);

  if ($this->{normalizeGroupName}) {
    $groupName = $this->normalizeWikiName($groupName);
  }
  return 0 if $this->{excludeMap}{$groupName};

  # check for a rewrite rule
  my $foundRewriteRule = 0;
  my $oldGroupName = $groupName;
  foreach my $pattern (keys %{$this->{rewriteGroups}}) {
    my $subst = $this->{rewriteGroups}{$pattern};
    if ($groupName =~ /^(?:$pattern)$/) {
      my $arg1 = $1;
      my $arg2 = $2;
      my $arg3 = $3;
      my $arg4 = $4;
      my $arg5 = $5;
      $arg1 = '' unless defined $arg1;
      $arg2 = '' unless defined $arg2;
      $arg3 = '' unless defined $arg3;
      $arg4 = '' unless defined $arg4;
      $subst =~ s/\$1/$arg1/g;
      $subst =~ s/\$2/$arg2/g;
      $subst =~ s/\$3/$arg3/g;
      $subst =~ s/\$4/$arg4/g;
      $subst =~ s/\$5/$arg5/g;
      $groupName = $subst;
      $foundRewriteRule = 1;
      #writeDebug("rewriting '$oldGroupName' to '$groupName' using rule $pattern");
      last;
    }
  }

  if (!$this->{mergeGroups} && defined($groupNames->{$groupName})) {
    writeWarning("$dn clashes with group $groupNames->{$groupName} on $groupName");
    return 0;
  }

  my $loginName = $this->{caseSensitiveLogin} ? $groupName : lc($groupName);
  if (defined($data->{"U2W::$loginName"}) || defined($data->{"W2U::$groupName"})) {
    my $groupSuffix = '';
    if ($this->{normalizeGroupName}) {
      $groupSuffix = 'Group';
    } else {
      $groupSuffix = '_group';
    }
    #writeWarning("group $dn clashes with user $groupName ... appending $groupSuffix");
    $groupName .= $groupSuffix;
  }

  # remember this
  $data->{"DN2U::$dn"} = $groupName;
  $data->{"U2DN::$groupName"} = $dn;

  # cache groupIds
  my $groupId = $entry->get_value($this->{primaryGroupAttribute});
  if ($groupId) {
    $this->{_groupId}{$groupId} = $groupName;
  }

  # fetch all members of this group
  foreach my $member ($entry->get_value($this->{memberAttribute})) {
    next unless $member;
    $member =~ s/^\s+//o;
    $member =~ s/\s+$//o;
    ($member =~ /mgrady|branson/) && writeDebug("cache group - adding member='$member' for groupName='$groupName', dn=$dn");
    $this->{_groups}{$groupName}{$member} = 1;    # delay til all groups have been fetched
  }

  # fetch all inner groups of this group
  foreach my $innerGroup ($entry->get_value($this->{innerGroupAttribute})) {
    next unless $innerGroup;
    $innerGroup =~ s/^\s+//o;
    $innerGroup =~ s/\s+$//o;
    $this->{_groups}{$groupName}{$innerGroup} = 1;    # delay til all groups have been fetched
  }

  # store it
  #writeDebug("adding groupName='$groupName', dn=$dn");

  $groupNames->{$groupName} = 1;

  return 1;
}

=pod

---++ cacheShibGroupsForPerson($loginName, $data, $groupNames) -> $boolean

store any Shib-delivered groups to our "memory" cache, 
actually just entries in the ShibLdapContrib object {data}.

returns true if new records have been created

=cut

sub cacheShibGroupsForPerson {
  my ($this, $loginName, $data) = @_;
  $data ||= $this->{data};

  # Add this user's groups, from the appropriate Shib attributes, to the Group cache,
  # if the ShibMapGroups setting is 'on'
  unless ($this->{shibMapGroups} && $ENV{$this->{shibGroupAttribute}}) {
      writeDebug("caching of Shib Groups is turned off, or no environment variable set");
      return 0;
  }

  my %groupNames = map { $_ => 1 } @{$this->getGroupNames($data)};
  my $groupNames = \%groupNames;
  my %shibGroups = {}; 
  writeDebug("caching Shib groups for $loginName");

  # get group values
  my $groupValue = $ENV{$this->{shibGroupAttribute}};
  $groupValue = Foswiki::Sandbox::untaintUnchecked($groupValue);

  unless ($groupValue) {
    writeDebug("no groupValue for $loginName ... skipping");
    return 0;
  }

  # Need to split values on semicolon, but keep semicolon escaped with backslash
  # Found this approach on Stack Exchange
  my @groups = map { s/\\([;])/$1/g; $_; } ($groupValue =~ /(?:^|;) ((?:\\[;] | [^;])*)/gx);

  foreach my $groupName (@groups) { 
    $groupName =~ s/^\s+//o;
    $groupName =~ s/\s+$//o;
    $groupName = $this->fromLdapCharSet($groupName);

    if ($this->{shibNormalizeGroupName}) {
      $groupName = $this->normalizeWikiName($groupName);
    }
    return 0 if $this->{excludeMap}{$groupName};

    writeDebug("found group '$groupName' for $loginName");

    # check for a rewrite rule
    my $foundRewriteRule = 0;
    my $oldGroupName = $groupName;
    foreach my $pattern (keys %{$this->{shibRewriteGroupNames}}) {
      my $subst = $this->{shibRewriteGroupNames}{$pattern};
      if ($groupName =~ /^(?:$pattern)$/) {
        my $arg1 = $1;
        my $arg2 = $2;
        my $arg3 = $3;
        my $arg4 = $4;
        my $arg5 = $5;
        $arg1 = '' unless defined $arg1;
        $arg2 = '' unless defined $arg2;
        $arg3 = '' unless defined $arg3;
        $arg4 = '' unless defined $arg4;
        $subst =~ s/\$1/$arg1/g;
        $subst =~ s/\$2/$arg2/g;
        $subst =~ s/\$3/$arg3/g;
        $subst =~ s/\$4/$arg4/g;
        $subst =~ s/\$5/$arg5/g;
        $groupName = $subst;
        $foundRewriteRule = 1;
        writeDebug("rewriting '$oldGroupName' to '$groupName' using rule $pattern");
        last;
      }
    }

    if (!$this->{shibMergeWithLdapGroups} && defined($groupNames->{$groupName})) {
      writeWarning("$loginName group clashes with group $groupNames->{$groupName} on $groupName");
      return 0;
    }

    my $loginNameGroup = $this->{caseSensitiveLogin} ? $groupName : lc($groupName);
    if (defined($data->{"U2W::$loginNameGroup"}) || defined($data->{"W2U::$groupName"})) {
      my $groupSuffix = '';
      if ($this->{shibNormalizeGroupName}) {
        $groupSuffix = 'Group';
      } else {
        $groupSuffix = '_group';
      }
      #writeWarning("group $groupName for $loginName clashes with user $groupName ... appending $groupSuffix");
      $groupName .= $groupSuffix;
    }

    # store it
    writeDebug("adding groupName='$groupName' to Shib cache");
    $this->{shibUserGroupNames}{$groupName} = 1;
    $shibGroups{$groupName} = 1;
  }

  # remember list of all groups
  $this->{shibUserAllGroups} = join(',', sort keys %shibGroups);

  return 1;
}

=pod 

---++ normalizeWikiName($name) -> $string

normalizes a string to form a proper <nop>WikiName

=cut

sub normalizeWikiName {
  my ($this, $name) = @_;

  $name = transliterate($name);

  my $wikiName = '';

  # first, try without forcing each part to be lowercase
  foreach my $part (split(/[^$Foswiki::regex{mixedAlphaNum}]/, $name)) {
    $wikiName .= ucfirst($part);
  }

  # if it isn't a valid WikiWord and there's no homepage of that name yet, then try more agressively to
  # create a proper WikiName
  if (!Foswiki::Func::isValidWikiWord($wikiName) && !Foswiki::Func::topicExists($Foswiki::cfg{UsersWebName}, $wikiName)) {
    $wikiName = '';
    foreach my $part (split(/[^$Foswiki::regex{mixedAlphaNum}]/, $name)) {
      $wikiName .= ucfirst(lc($part));
    }
  }

  return $wikiName;
}

=pod 

---++ shibNormalizeLoginName($name) -> $string

normalizes a string to form a proper login, without removing the @scope
and without applying the usual Foswiki LoginNameFilterIn

=cut

sub shibNormalizeLoginName {
  my ($this, $name) = @_;

  $name = transliterate($name);
  writeDebug("normalizeLoginName to $name");

  return $name;
}

=pod

---++ normalizeLoginName($name) -> $string

normalizes a string to form a proper login

=cut

sub normalizeLoginName {
  my ($this, $name) = @_;

  $name =~ s/@.*$//o;    # remove REALM

  $name = transliterate($name);
  $name =~ s/[^$Foswiki::cfg{LoginNameFilterIn}]//;

  return $name;
}

=pod

---++ transliterate($string) -> $string

transliterate some essential utf8 chars to a common replacement
in latin1 encoding. the list above is not exhaustive.

use http://www.ltg.ed.ac.uk/~richard/utf-8.html to add more recodings

=cut

sub transliterate {
  my $string = shift;

  if ($Foswiki::cfg{Site}{CharSet} =~ /^utf-?8$/i) {
    $string =~ s/\xc3\xa0/a/go;    # a grave
    $string =~ s/\xc3\xa1/a/go;    # a acute
    $string =~ s/\xc3\xa2/a/go;    # a circumflex
    $string =~ s/\xc3\xa3/a/go;    # a tilde
    $string =~ s/\xc3\xa4/ae/go;   # a uml
    $string =~ s/\xc3\xa5/a/go;    # a ring above
    $string =~ s/\xc3\xa6/ae/go;   # ae
    $string =~ s/\xc4\x85/a/go;    # a ogonek

    $string =~ s/\xc3\x80/A/go;    # A grave
    $string =~ s/\xc3\x81/A/go;    # A acute
    $string =~ s/\xc3\x82/A/go;    # A circumflex
    $string =~ s/\xc3\x83/A/go;    # A tilde
    $string =~ s/\xc3\x84/Ae/go;   # A uml
    $string =~ s/\xc3\x85/A/go;    # A ring above
    $string =~ s/\xc3\x86/AE/go;   # AE
    $string =~ s/\xc4\x84/A/go;    # A ogonek

    $string =~ s/\xc3\xa7/c/go;    # c cedille
    $string =~ s/\xc4\x87/c/go;    # c acute
    $string =~ s/\xc4\x8d/c/go;    # c caron
    $string =~ s/\xc3\x87/C/go;    # C cedille
    $string =~ s/\xc4\x86/C/go;    # C acute
    $string =~ s/\xc4\x8c/C/go;    # C caron

    $string =~ s/\xc4\x8f/d/go;    # d caron
    $string =~ s/\xc4\x8e/D/go;    # D caron

    $string =~ s/\xc3\xa8/e/go;    # e grave
    $string =~ s/\xc3\xa9/e/go;    # e acute
    $string =~ s/\xc3\xaa/e/go;    # e circumflex
    $string =~ s/\xc3\xab/e/go;    # e uml
    $string =~ s/\xc4\x9b/e/go;    # e caron

    $string =~ s/\xc4\x99/e/go;    # e ogonek
    $string =~ s/\xc4\x98/E/go;    # E ogonek
    $string =~ s/\xc4\x9a/E/go;    # E caron

    $string =~ s/\xc3\xb2/o/go;    # o grave
    $string =~ s/\xc3\xb3/o/go;    # o acute
    $string =~ s/\xc3\xb4/o/go;    # o circumflex
    $string =~ s/\xc3\xb5/o/go;    # o tilde
    $string =~ s/\xc3\xb6/oe/go;   # o uml
    $string =~ s/\xc3\xb8/o/go;    # o stroke

    $string =~ s/\xc3\xb3/o/go;    # o acute
    $string =~ s/\xc3\x93/O/go;    # O acute

    $string =~ s/\xc3\x92/O/go;    # O grave
    $string =~ s/\xc3\x93/O/go;    # O acute
    $string =~ s/\xc3\x94/O/go;    # O circumflex
    $string =~ s/\xc3\x95/O/go;    # O tilde
    $string =~ s/\xc3\x96/Oe/go;   # O uml

    $string =~ s/\xc3\xb9/u/go;    # u grave
    $string =~ s/\xc3\xba/u/go;    # u acute
    $string =~ s/\xc3\xbb/u/go;    # u circumflex
    $string =~ s/\xc3\xbc/ue/go;   # u uml
    $string =~ s/\xc5\xaf/u/go;    # u ring above

    $string =~ s/\xc3\x99/U/go;    # U grave
    $string =~ s/\xc3\x9a/U/go;    # U acute
    $string =~ s/\xc3\x9b/U/go;    # U circumflex
    $string =~ s/\xc3\x9c/Ue/go;   # U uml
    $string =~ s/\xc5\xae/U/go;    # U ring above

    $string =~ s/\xc5\x99/r/go;    # r caron
    $string =~ s/\xc5\x98/R/go;    # R caron

    $string =~ s/\xc3\x9f/ss/go;   # sharp s
    $string =~ s/\xc5\x9b/s/go;    # s acute
    $string =~ s/\xc5\xa1/s/go;    # s caron
    $string =~ s/\xc5\x9a/S/go;    # S acute
    $string =~ s/\xc5\xa0/S/go;    # S caron

    $string =~ s/\xc5\xa5/t/go;    # t caron
    $string =~ s/\xc5\xa4/T/go;    # T caron

    $string =~ s/\xc3\xb1/n/go;    # n tilde
    $string =~ s/\xc5\x84/n/go;    # n acute
    $string =~ s/\xc5\x88/n/go;    # n caron
    $string =~ s/\xc5\x83/N/go;    # N acute
    $string =~ s/\xc5\x87/N/go;    # N caron

    $string =~ s/\xc3\xbe/y/go;    # y acute
    $string =~ s/\xc3\xbf/y/go;    # y uml

    $string =~ s/\xc3\xac/i/go;    # i grave
    $string =~ s/\xc3\xab/i/go;    # i acute
    $string =~ s/\xc3\xac/i/go;    # i circumflex
    $string =~ s/\xc3\xad/i/go;    # i uml

    $string =~ s/\xc5\x82/l/go;    # l stroke
    $string =~ s/\xc4\xbe/l/go;    # l caron
    $string =~ s/\xc5\x81/L/go;    # L stroke
    $string =~ s/\xc4\xbd/L/go;    # L caron

    $string =~ s/\xc5\xba/z/go;    # z acute
    $string =~ s/\xc5\xb9/Z/go;    # Z acute
    $string =~ s/\xc5\xbc/z/go;    # z dot
    $string =~ s/\xc5\xbb/Z/go;    # Z dot
    $string =~ s/\xc5\xbe/z/go;    # z caron
    $string =~ s/\xc5\xbd/Z/go;    # Z caron
  } else {
    $string =~ s/\xe0/a/go;        # a grave
    $string =~ s/\xe1/a/go;        # a acute
    $string =~ s/\xe2/a/go;        # a circumflex
    $string =~ s/\xe3/a/go;        # a tilde
    $string =~ s/\xe4/ae/go;       # a uml
    $string =~ s/\xe5/a/go;        # a ring above
    $string =~ s/\xe6/ae/go;       # ae
    $string =~ s/\x01\x05/a/go;    # a ogonek

    $string =~ s/\xc0/A/go;        # A grave
    $string =~ s/\xc1/A/go;        # A acute
    $string =~ s/\xc2/A/go;        # A circumflex
    $string =~ s/\xc3/A/go;        # A tilde
    $string =~ s/\xc4/Ae/go;       # A uml
    $string =~ s/\xc5/A/go;        # A ring above
    $string =~ s/\xc6/AE/go;       # AE
    $string =~ s/\x01\x04/A/go;    # A ogonek

    $string =~ s/\xe7/c/go;        # c cedille
    $string =~ s/\x01\x07/C/go;    # c acute
    $string =~ s/\xc7/C/go;        # C cedille
    $string =~ s/\x01\x06/c/go;    # C acute

    $string =~ s/\xe8/e/go;        # e grave
    $string =~ s/\xe9/e/go;        # e acute
    $string =~ s/\xea/e/go;        # e circumflex
    $string =~ s/\xeb/e/go;        # e uml
    $string =~ s/\x01\x19/e/go;    # e ogonek
    $string =~ s/\xc4\x18/E/go;    # E ogonek

    $string =~ s/\xf2/o/go;        # o grave
    $string =~ s/\xf3/o/go;        # o acute
    $string =~ s/\xf4/o/go;        # o circumflex
    $string =~ s/\xf5/o/go;        # o tilde
    $string =~ s/\xf6/oe/go;       # o uml
    $string =~ s/\xf8/oe/go;       # o stroke

    $string =~ s/\xd3/o/go;        # o acute
    $string =~ s/\xf3/O/go;        # O acute

    $string =~ s/\xd2/O/go;        # O grave
    $string =~ s/\xd3/O/go;        # O acute
    $string =~ s/\xd4/O/go;        # O circumflex
    $string =~ s/\xd5/O/go;        # O tilde
    $string =~ s/\xd6/Oe/go;       # O uml

    $string =~ s/\xf9/u/go;        # u grave
    $string =~ s/\xfa/u/go;        # u acute
    $string =~ s/\xfb/u/go;        # u circumflex
    $string =~ s/\xfc/ue/go;       # u uml

    $string =~ s/\xd9/U/go;        # U grave
    $string =~ s/\xda/U/go;        # U acute
    $string =~ s/\xdb/U/go;        # U circumflex
    $string =~ s/\xdc/Ue/go;       # U uml

    $string =~ s/\xdf/ss/go;       # sharp s
    $string =~ s/\x01\x5b/s/go;    # s acute
    $string =~ s/\x01\x5a/S/go;    # S acute

    $string =~ s/\xf1/n/go;        # n tilde
    $string =~ s/\x01\x44/n/go;    # n acute
    $string =~ s/\x01\x43/N/go;    # N acute

    $string =~ s/\xfe/y/go;        # y acute
    $string =~ s/\xff/y/go;        # y uml

    $string =~ s/\xec/i/go;        # i grave
    $string =~ s/\xed/i/go;        # i acute
    $string =~ s/\xee/i/go;        # i circumflex
    $string =~ s/\xef/i/go;        # i uml

    $string =~ s/\x01\x42/l/go;    # l stroke
    $string =~ s/\x01\x41/L/go;    # L stroke

    $string =~ s/\x01\x7a/z/go;    # z acute
    $string =~ s/\x01\x79/Z/go;    # Z acute
    $string =~ s/\x01\x7c/z/go;    # z dot
    $string =~ s/\x01\x7b/Z/go;    # Z dot
  }

  return $string;
}

=pod

---++ getGroupNames($data) -> @array

Returns a list of known group names.

=cut

sub getGroupNames {
  my ($this, $data) = @_;

  #writeDebug("called getGroupNames()");

  $data ||= $this->{data};

  my $groupNames = Foswiki::Sandbox::untaintUnchecked($data->{GROUPS}) || '';
  my @groupNames = split(/\s*,\s*/, $groupNames);

  # add current Shib groups to the end of the array
  if ($this->{shibUserAllGroups}) {
	my @shibGroupNames = split(/\s*,\s*/, $this->{shibUserAllGroups});
	push (@groupNames, @shibGroupNames);
  }

  return \@groupNames;
}

=pod

---++ isGroup($wikiName, $data) -> $boolean

check if a given user is an ldap group actually

=cut

sub isGroup {
  my ($this, $wikiName, $data) = @_;

  #writeDebug("called isGroup($wikiName)");
  $data ||= $this->{data};

  return undef if $this->{excludeMap}{$wikiName};
  return 1 if defined($data->{"GROUPS::$wikiName"});
  return 0 if defined($data->{"W2U::$wikiName"});

  my $loginName = $this->{caseSensitiveLogin} ? $wikiName : lc($wikiName);
  return 0 if defined($data->{"U2W::$loginName"});

  # check to see if this name is a Shib-delivered group name
  return 1 if defined($this->{shibUserGroupNames}{$wikiName});  

  unless ($this->{preCache}) {
    $this->checkCacheForGroupName($wikiName, $data);
    return 1 if defined($data->{"GROUPS::$wikiName"});
  }

  return undef;
}

=pod

---++ getEmails($loginName, $data) -> @emails

fetch emails from LDAP

=cut

sub getEmails {
  my ($this, $loginName, $data) = @_;
  my @emails;

  $loginName = lc($loginName) unless $this->{caseSensitiveLogin};
  $data ||= $this->{data};
  
  # check first to see if this matches for the current Shib user
  if ($this->{shibUserLoginName} eq $loginName) {
	@emails = split(/\s*,\s*/, $this->{shibUserEmailAddresses}); 
	return \@emails;
  }
  
  $this->checkCacheForLoginName($loginName, $data) unless $this->{preCache};

  my $emails = Foswiki::Sandbox::untaintUnchecked($data->{"U2EMAIL::" . $loginName}) || '';
  @emails = split(/\s*,\s*/, $emails);
  return \@emails;
}

=pod

---++ getLoginOfEmail($email, $data) \@users

get all users matching a given email address

=cut

sub getLoginOfEmail {
  my ($this, $email, $data) = @_;

  $data ||= $this->{data};

  my $loginNames = Foswiki::Sandbox::untaintUnchecked($data->{"EMAIL2U::" . $email}) || '';
  my @loginNames = split(/\s*,\s*/, $loginNames);

  # check to see if this matches for the current Shib user
  if ($this->{shibUserEmailsForLoginName}{$email}) {
	push(@loginNames, $this->{shibUserLoginName});
  }

  return \@loginNames;

}

=pod

---++ getGroupMembers($groupName, $data) -> \@array

=cut

sub getGroupMembers {
  my ($this, $groupName, $data) = @_;
  return undef if $this->{excludeMap}{$groupName};

  #writeDebug("called getGroupMembers $groupName");

  $data ||= $this->{data};

  unless ($this->{preCache}) {

    # Make sure that the group is in the cache. This will cause the addition of the group to the cache if it exists in LDAP
    return undef unless $this->isGroup($groupName, $data);
  }

  my $members = Foswiki::Sandbox::untaintUnchecked($data->{"GROUPS::$groupName"}) || '';
  my @members = split(/\s*,\s*/, $members);

  # add current Shib user loginName if member of this group
  if ($this->{shibUserGroupNames}{$groupName}) {
	push (@members, $this->{shibUserLoginName});
  }

  return \@members;
}

=pod

---++ isGroupMember($loginName, $groupName, $data) -> $boolean

check if a given user is member of an ldap group

=cut

sub isGroupMember {
  my ($this, $loginName, $groupName, $data) = @_;

  $data ||= $this->{data};

  # check first to see if this user matches the current Shib user, and if so,
  # if the group matches one of their groups
  if ($this->{shibUserLoginName} eq $loginName) {
	return 1 if defined($this->{shibUserGroupNames}{$groupName});
  }
   
  unless ($this->{preCache}) {
    # We need to make sure both user and group are in the cache. These calls will trigger LDAP lookups if appropriate.
    return 0 unless $this->checkCacheForLoginName($loginName, $data);
    return 0 unless $this->isGroup($groupName, $data);
  }

  my $members = $data->{"GROUPS::$groupName"} || '';
  return ($members =~ /\b$loginName\b/) ? 1 : 0;
}

=pod 

---++ getWikiNameOfLogin($loginName, $data) -> $wikiName

returns the wikiName of a loginName or undef if it does not exist

=cut

sub getWikiNameOfLogin {
  my ($this, $loginName, $data) = @_;

  #writeDebug("called getWikiNameOfLogin($loginName)");

  $loginName = lc($loginName) unless $this->{caseSensitiveLogin};
  $data ||= $this->{data};

  # check to see if this matches current Shib user first
  if ($this->{shibUserLoginName} eq $loginName) {
	return $this->{shibUserWikiName};
  }

  unless ($this->{preCache}) {
    # Make sure the user has been retreived from LDAP
    $this->checkCacheForLoginName($loginName, $data);
  }
  
  return Foswiki::Sandbox::untaintUnchecked($data->{"U2W::$loginName"});
}

=pod 

---++ getLoginOfWikiName($wikiName, $data) -> $loginName

returns the loginName of a wikiName or undef if it does not exist

=cut

sub getLoginOfWikiName {
  my ($this, $wikiName, $data) = @_;

  #writeDebug("called getLoginOfWikiName($wikiName)");
 
  $data ||= $this->{data};

  # check to see if this matches current Shib user first
  if ($this->{shibUserWikiName} eq $wikiName) {
	return $this->{shibUserLoginName};
  }
  
  my $loginName = Foswiki::Sandbox::untaintUnchecked($data->{"W2U::$wikiName"});

  unless ($loginName) {
    my $alias = $this->{wikiNameAliases}{$wikiName};
    $loginName = Foswiki::Sandbox::untaintUnchecked($data->{"W2U::$alias"})
      if defined($alias);
  }

  return $loginName;
}

=pod 

---++ getAllWikiNames($data) -> \@array

returns a list of all known wikiNames

=cut

sub getAllWikiNames {
  my ($this, $data) = shift;

  $data ||= $this->{data};

  my $wikiNames = Foswiki::Sandbox::untaintUnchecked($data->{WIKINAMES}) || '';
  my @wikiNames = split(/\s*,\s*/, $wikiNames);

  # add current Shib user wikiName to the end of the array
  push(@wikiNames, $this->{shibUserWikiName});
  
  return \@wikiNames;
}

=pod 

---++ getAllLoginNames($data) -> \@array

returns a list of all known loginNames

=cut

sub getAllLoginNames {
  my ($this, $data) = @_;

  $data ||= $this->{data};

  my $loginNames = Foswiki::Sandbox::untaintUnchecked($data->{LOGINNAMES}) || '';
  my @loginNames = split(/\s*,\s*/, $loginNames);

  # add current Shib user loginName to the end of the array
  push(@loginNames, $this->{shibUserLoginName});

  return \@loginNames;
}

=pod 

---++ getDnOfLogin($loginName, $data) -> $dn

returns the Distinguished Name of the LDAP record of the given name

=cut

sub getDnOfLogin {
  my ($this, $loginName, $data) = @_;

  return unless $loginName;

  $loginName = lc($loginName) unless $this->{caseSensitiveLogin};
  $data ||= $this->{data};

  return Foswiki::Sandbox::untaintUnchecked($data->{"U2DN::$loginName"});
}

=pod 

---++ getDnOfWikiName($wikiName, $data) -> $dn

returns the Distinguished Name of the LDAP record of the given name

=cut

sub getDnOfWikiName {
  my ($this, $wikiName, $data) = @_;

  return unless $wikiName;

  $data ||= $this->{data};

  my $loginName = Foswiki::Sandbox::untaintUnchecked($data->{"W2U::$wikiName"});
  return unless $loginName;

  return Foswiki::Sandbox::untaintUnchecked($data->{"U2DN::$loginName"});
}

=pod 

---++ getWikiNameOfDn($dn, $data) -> $wikiName

returns the wikiName used by a given Distinguished Name; reverse of getDnOfWikiName()

=cut

sub getWikiNameOfDn {
  my ($this, $dn, $data) = @_;

  return unless $dn;

  $data ||= $this->{data};

  my $loginName = Foswiki::Sandbox::untaintUnchecked($data->{"DN2U::$dn"});
  return unless $loginName;

  return Foswiki::Sandbox::untaintUnchecked($data->{"U2W::$loginName"});
}

=pod 

---++ changePassword($loginName, $newPassword, $oldPassword) -> $boolean

=cut

sub changePassword {
  my ($this, $loginName, $newPassword, $oldPassword) = @_;

  return undef
    unless $this->{allowChangePassword} && defined($oldPassword) && $oldPassword ne '1';

  my $dn = $this->getDnOfLogin($loginName);
  return undef unless $dn;

  return undef unless $this->connect($dn, $oldPassword);

  my $msg = $this->{ShibLdap}->set_password(
    oldpasswd => $oldPassword,
    newpasswd => $newPassword
  );

  my $errorCode = $this->checkError($msg);

  if ($errorCode != LDAP_SUCCESS) {
    writeWarning("error in changePassword: " . $this->getError());
    return undef;
  }

  return 1;
}

=pod

---++ checkCacheForLoginName($loginName, $data) -> $boolean

grant that the current loginName is cached. If not, it will download the LDAP
record for this specific user and update the LDAP cache with this single record.

This happens when the user is authenticated externally, e.g. using apache's
mod_authz_ldap or some other SSO, and the internal cache 
is not yet updated. It is completely updated regularly on a specific time
interval (default every 24h). See the LdapContrib settings.

=cut

sub checkCacheForLoginName {
  my ($this, $loginName, $data) = @_;

  return 0 unless ($loginName);

  writeDebug("called checkCacheForLoginName($loginName)");

  $loginName = lc($loginName) unless $this->{caseSensitiveLogin};
  $data ||= $this->{data};

  return 1 if $data->{"U2W::$loginName"};

  # check to see if this matches current Shib user, who will only be in "Shib memory cache"
  if ($this->{shibUserLoginName} eq $loginName) {
	return 1;
  }
  
  # If we are not in precache mode we need to check if the user has not yet been unsuccessfully lookedup in LDAP
  # To avoid excessive useless queries for the same non existing user
  unless ($this->{preCache}) {
    if ($this->isUnknownUser($loginName, $data)) {
      return 0;
    }
  }

  # update cache selectively
  writeDebug("$loginName is unknown, need to refresh part of the ldap cache");

  my $entry = $this->getAccount($loginName);

  unless ($entry) {
    writeWarning("oops, no result looking for user $loginName in LDAP");
    $this->addIgnoredUser($loginName, $data);
  } else {
    # merge this user record

    my %wikiNames = map { $_ => 1 } @{$this->getAllWikiNames($data)};
    my %loginNames = map { $_ => 1 } @{$this->getAllLoginNames($data)};
    $this->cacheUserFromEntry($entry, $data, \%wikiNames, \%loginNames);

    $data->{WIKINAMES} = join(',', keys %wikiNames);
    $data->{LOGINNAMES} = join(',', keys %loginNames);

    return 1;
  }

  return 0;
}

=pod

---++ removeGroupFromCache($groupName, $data) -> $boolean

Remove a group from the cache

=cut

sub removeGroupFromCache {
  my ($this, $groupName, $data) = @_;

  return 0 unless defined $groupName;

  $data ||= $this->{data};

  my %groupNames = map { $_ => 1 } @{$this->getGroupNames($data)};
  my $dn = $this->getDnOfLogin($groupName, $data);

  delete $groupNames{$groupName};
  delete $data->{"GROUPS::$groupName"};
  delete $data->{"GROUP2UNCACHEDMEMBERSDN::$groupName"};
  delete $data->{"U2DN::$groupName"};
  delete $data->{"DN2U::$dn"} if defined $dn;

  $data->{GROUPS} = join(',', keys %groupNames);

  return 1;
}

=pod

---++ removeUserFromCache($wikiName, $data) -> $boolean

removes a wikiName from the cache

=cut

sub removeUserFromCache {
  my ($this, $wikiName, $data) = @_;

  return 0 unless defined $wikiName;

  $data ||= $this->{data};

  my %wikiNames = map { $_ => 1 } @{$this->getAllWikiNames($data)};
  my %loginNames = map { $_ => 1 } @{$this->getAllLoginNames($data)};
  my $loginName = $this->getLoginOfWikiName($wikiName);
  my $dn = $this->getDnOfLogin($loginName, $data);

  delete $loginNames{$loginName};
  delete $wikiNames{$wikiName};
  delete $data->{"U2W::$loginName"};
  delete $data->{"W2U::$wikiName"};
  delete $data->{"DN2U::$dn"};
  delete $data->{"U2DN::$loginName"};
  delete $data->{"U2EMAIL::$loginName"};

  foreach my $email (@{$this->getEmails($loginName, $data)}) {
    my %emails = map { $_ => 1 } split(/\s*,\s*/, $data->{"EMAIL2U::$email"});
    delete $emails{$loginName};
    $data->{"EMAIL2U::$email"} = join(',', sort keys %emails);
  }

  $data->{LOGINNAMES} = join(',', keys %loginNames);
  $data->{WIKINAMES} = join(',', keys %wikiNames);

}

=begin text

---++ renameWikiName($loginName, $oldWikiName, $newWikiName) 

assigns the new !WikiName to the given login

=cut

sub renameWikiName {
  my ($this, $loginName, $oldWikiName, $newWikiName, $data) = @_;

  $data ||= $this->{data};

  #writeDebug("renameWikiName($loginName, $oldWikiName, $newWikiName)");

  if (defined $data->{"W2U::$oldWikiName"}) {
    delete $data->{"W2U::$oldWikiName"};

    $data->{"U2W::$loginName"} = $newWikiName;
    $data->{"W2U::$newWikiName"} = $loginName;
    return 1;
  }

  #writeWarning("oldWikiName=$oldWikiName not found in cache");
  return 0;
}

=pod 

---++ addIgnoredUser($loginName, $data) -> \@array

Insert a new user in the list of unknown users that should not be lookedup in LDAP

=cut

sub addIgnoredUser {
  my ($this, $loginName, $data) = @_;

  $data ||= $this->{data};

  $data->{UNKWNUSERS} .= ',' . $loginName
    unless $this->isUnknownUser($loginName, $data);
}

=pod 

---++++ isUnknownUser($loginName, $data) -> $boolean

returns 1 if $loginName is an unknown user that should not be relookedup in LDAP

=cut

sub isUnknownUser {
  my ($this, $loginName, $data) = @_;

  $data ||= $this->{data};

  my $names = Foswiki::Sandbox::untaintUnchecked($data->{UNKWNUSERS}) || '';
  return $names =~ /\b$loginName\b/;
}

=pod 

---++ addIgnoredGroup($groupName, $data) -> \@array

Insert a new group in the list of unknown groups that should not be lookedup in LDAP

=cut

sub addIgnoredGroup {
  my ($this, $groupName, $data) = @_;

  $data ||= $this->{data};

  $data->{UNKWNGROUPS} .= ',' . $groupName
    unless $this->isUnknownGroup($groupName, $data);
}

=pod 

---++++ isUnknownGroup($groupName, $data) -> $boolean

returns 1 if $groupName is an unknown groups that should not be relookedup in LDAP

=cut

sub isUnknownGroup {
  my ($this, $groupName, $data) = @_;

  $data ||= $this->{data};

  my $wikiNames = Foswiki::Sandbox::untaintUnchecked($data->{UNKWNGROUPS}) || '';
  return $wikiNames =~ /\b$groupName\b/;
}

=pod

---++ checkCacheForGroupName($groupName, $data) -> $boolean

grant that the current groupName is cached. If not, it will download the LDAP
record for this specific group and its subgroups and update the LDAP cache with the retreived records.

This happens when the precache mode is off. See the ShibLdapContrib settings.

=cut

sub checkCacheForGroupName {
  my ($this, $groupName, $data) = @_;

  #writeDebug("called checkCacheForGroupName($groupName)");

  $data ||= $this->{data};

  # Skip lookup if group was already not found in LDAP since last cache expiration
  unless ($this->{preCache}) {
    if ($this->isUnknownGroup($groupName, $data)) {
      return 0;
    }
  }

  # update cache selectively
  #writeDebug("group $groupName is unknown, need to refresh part of the ldap cache");

  my $entry = $this->getGroup($groupName);
  unless ($entry) {

    writeWarning("oops, no result looking for group $groupName in LDAP");
    $this->addIgnoredGroup($groupName, $data);
    return 0;
  } else {

    # merge this group record
    my %groupNames = map { $_ => 1 } @{$this->getGroupNames($data)};

    $this->cacheGroupFromEntry($entry, $data, \%groupNames);

    # remember list of all groups
    $data->{GROUPS} = join(',', sort keys %groupNames);

    # check for primary group membership
    if ($this->{_primaryGroup}) {
      foreach my $groupId (keys %{$this->{_primaryGroup}}) {
        my $currentGroupName = $this->{_groupId}{$groupId};

        if (defined $currentGroupName && $groupName eq $currentGroupName) {
          foreach my $member (keys %{$this->{_primaryGroup}{$groupId}}) {

            #writeDebug("adding $member to its primary group $currentGroupName");
            $this->{_groups}{$currentGroupName}{$member} = 1;
          }
        }
      }
    }

    # assert group members to data store
    my %members = ();
    my %uncachedMembersDn = ();

    foreach my $member (keys %{$this->{_groups}{$groupName}}) {

      # groups may store DNs to members instead of a memberUid, in this case we
      # have to lookup the corresponding loginAttribute
      if ($this->{memberIndirection}) {

        #writeDebug("following indirection for $member");

        my $memberName = $data->{"DN2U::$member"};
        if ($memberName) {
          $members{$memberName} = 1;
        } else {

          # Recursive check for groups when not in precache mode
          if (!$this->{preCache} && $member =~ /$this->{groupBase}/i) {
            my $innerGroupName = $member;
            $innerGroupName =~ s/$this->{groupBase}//o;
            $innerGroupName =~ s/$this->{groupAttribute}=//oi;
            $innerGroupName =~ s/^,+//o;
            $innerGroupName =~ ($this->{UserScope} eq 'sub' || $this->{GroupAttribute} eq 'sub') ? s/,.*$//o : s/,+$//o;

            # Smell: this may not be reliable and may work only with membersindirection. TO CHECK
            if ($innerGroupName ne "" && $this->isGroup($innerGroupName, $data)) {
              $members{$innerGroupName} = 1;
              next;
            }
          }

          #writeWarning("oops, $member not found, but member of $groupName");
          $uncachedMembersDn{$member} = 1;
        }
      } else {
        $members{$member} = 1;
      }
    }

    $data->{"GROUPS::$groupName"} = join(',', sort keys %members);

    if ($this->{memberIndirection}) {
      $data->{"GROUP2UNCACHEDMEMBERSDN::$groupName"} = join(';', keys %uncachedMembersDn);
    }

    undef $this->{_groups}{$groupName};
    undef $this->{_groups};

    return 1;
  }
}

=pod

---++ getGroup($groupName) -> Net::LDAP::Entry object

Fetches a group entry from the database and returns a Net::LDAP::Entry
object on success and undef otherwise. Note, the group name is match against
the attribute defined in $ldap->{groupAttribute}. Account records are 
search using $ldap->{groupFilter} in the subtree defined by $ldap->{groupBase}.

=cut

sub getGroup {
  my ($this, $groupName) = @_;

  #writeDebug("called getGroup($groupName)");
  return undef if $this->{excludeMap}{$groupName};

  my $filter = '(&(' . $this->{groupFilter} . ')(' . $this->{groupAttribute} . '=' . $groupName . '))';
  my $entry;

  foreach my $groupBase (@{$this->{groupBase}}) {
    my $msg = $this->search(
      filter => $filter,
      base => $groupBase,
      callback => sub {
        my (undef, $result) = @_;
        return unless defined $result;
        $entry = $result;
      }
    );
    return $entry if $entry;
  }

  $this->{error} = 'Group invalid';
  #writeDebug($this->{error});

  return;
}

=pod

---++ fromLdapCharSet($string) -> $string

decode strings coming from ldap 

=cut

sub fromLdapCharSet {
  my ($this, $string) = @_;

  my $ldapCharSet = $Foswiki::cfg{ShibLdap}{CharSet} || 'utf-8';
  return Encode::decode($ldapCharSet, $string);
}

=begin text

---++ fromSiteCharSet($string) -> $string

encode strings coming from the site to be used talking to the ldap directory

=cut

sub fromSiteCharSet {
  my ($this, $string) = @_;

  my $siteCharSet = $Foswiki::cfg{Site}{CharSet};

  return Encode::decode($siteCharSet, $string);
}

1;

